"""Main entry point for the pydantic model."""

import math
import random
from typing import List, Optional, Tuple

import pydantic
from typing_extensions import Literal

from optom_tools.utils import strip_decimal

from .exceptions import PrescriptionError
from .models import Add, BaseModel, HorizontalPrism, VerticalPrism


class Prescription(BaseModel):
    """The prescription module contains methods to deal with spectacle prescriptions.

    Examples:
        Typical use:
        >>> rx = Prescription(sphere=0, cylinder=-1, axis=180)
        >>> str(rx)
        'plano / -1.00 x 180'
    """

    sphere: float = 0
    cylinder: float = 0
    axis: float = 180
    add: Add = Add()
    intermediate_add: Add = Add(working_distance_cm=50)
    back_vertex_mm: float = 12.0

    vertical_prism: VerticalPrism = VerticalPrism()
    horizontal_prism: HorizontalPrism = HorizontalPrism()
    reading_vertical_prism: VerticalPrism = VerticalPrism()
    reading_horizontal_prism: HorizontalPrism = HorizontalPrism()
    extra_adds: List[Add] = []

    @pydantic.validator("axis")
    @classmethod
    def _axis_valid(cls, value: float) -> float:
        """Validate axis.

        Axis must be between 180 to 0 degrees.
        """
        if value > 180 or value < 0:
            raise PrescriptionError(
                value=value, message="Axis must be between 0 and 180 degrees"
            )
        return value

    @property
    def mean_sphere(self) -> float:
        """Provide mean sphere value of the prescription.

        Returns:
            (float): The Mean Sphere.
        """
        return self.sphere + (self.cylinder / 2)

    def __init__(self, *args, **kwargs):
        """Init method."""
        if len(args) >= 1:
            components = self._simple_parse_rx(args[0])
            kwargs["sphere"] = components[0]
            kwargs["cylinder"] = components[1]
            kwargs["axis"] = components[2]
        super().__init__(**kwargs)

    def transpose(self, flag: Optional[Literal["n", "p"]] = None) -> None:
        """Transpose prescription from positive to negative and vice versa.

        Flags, `'n'` and `'p'`, can be provided to force a negative or positive cylinder respectively.

        Args:
            (Optional[[Literal["n", "p"]]): Flag to force negative ('n') and positive ('p') cylindrical format. Defaults to `None`.

        Examples:
            Transposing a prescription as normal:
            >>> rx = Prescription("+1.00/-1.00x180")
            >>> rx.transpose()
            >>> str(rx)
            "plano / +1.00 x 90"
            >>> rx.transpose()
            >>> str(rx)
            "+1.00 / -1.00 x 180"

            Transposing a prescription with the 'n' flag:
            >>> rx = Prescription("+1.00/-1.00x180").transpose('n')
            >>> str(rx)
            "+1.00 / -1.00 x 180"
        """
        if flag is not None and flag not in ["n", "p"]:
            raise PrescriptionError(
                value=flag,
                message="Method transpose() only accepts 'n' and 'p' as input flags",
            )
        if (
            flag == "n"
            and self.cylinder > 0
            or flag == "p"
            and self.cylinder < 0
            or flag is None
            and self.cylinder != 0
        ):
            self.sphere = self.sphere + self.cylinder
            self.cylinder = -1 * self.cylinder
            new_axis = self.axis + 90
            if new_axis > 180:
                new_axis = new_axis - 180
            self.axis = new_axis

    def _simple_parse_rx(self, rx: str) -> Tuple[str, str, str]:
        """Parse rx for a simple input.

        And returns a tuple with sphere, cylinder and axis as strings.

        For example: simple input is '+1.00/-1.00x90'
        """
        # TODO: parse this '+1.00/-1.00x90 Add +2.00@40'
        rx_components = rx.split("/")
        if len(rx_components) > 2:
            raise PrescriptionError(
                value=rx_components, message="Only one '/' can be parsed."
            )
        sphere = rx_components[0]
        cylinder = "0"
        axis = "180"
        if len(rx_components) > 1:
            cylinder_all = rx_components[1]

            cyl_components = cylinder_all.split("x")
            cylinder = cyl_components[0]
            axis = cyl_components[1]

        if type(sphere) == str and sphere[0:2] == "pl":
            sphere = "0"

        return (sphere, cylinder, axis)

        # TODO: efficient Rx parser to be written.

    def parse(self, rx: str) -> BaseModel:
        """Parse a prescription in a more typical format.

        This is more familiar than setting a prescription using keyword arguments.

        Args:
            rx (str): The prescription as a string.

        Examples:
            Parsing a simple prescription:
            >>> rx = Prescription().parse("+1.00/-1.00x180")
            >>> rx.transpose()
            >>> str(rx)
            'pl / +1.00 x 90'
        """
        rx_tuple = self._simple_parse_rx(rx)
        self.sphere = float(rx_tuple[0])
        self.cylinder = float(rx_tuple[1])
        self.axis = float(rx_tuple[2])
        return self

    # TODO: need to distribute rx normally.
    def random(self, rx_range: float = 30, seed: Optional[int] = None) -> None:
        """Generate a random prescription.

        All results will produce negative cylinderical results. Use the transpose() method to return positive cylindrical results.

        Args:
            rx_range (float): The +/- range of prescription in dioptres.
            seed (Optional[int]): Seed for setting random.seed().

        Examples:
            Typical use:
            >>> rx = Prescription().random()
            >>> str(rx)
            '+1.00 / -1.00 x 173'
        """
        MEAN = -1
        STD = 1

        def _pdf(x: float, mean: float = MEAN, std: float = STD):
            """Point density function for weighting random selection of prescriptions.

            This is inplace because more extreme prescriptions like -30 are rare.
            """
            y = (x - mean) / std
            inter = (math.exp(-1 * y**2 / 2)) / math.sqrt(2 * math.pi)
            return inter / std

        rx_lst = [
            x / 100 for x in range(int(-1 * rx_range * 100), int(rx_range * 100), 25)
        ]
        weights = [_pdf(x) for x in rx_lst]
        if seed is not None:
            random.seed(seed)
        self.sphere = random.choices(rx_lst, weights=weights, k=1)[0]
        self.cylinder = abs(random.choices(rx_lst, weights=weights, k=1)[0]) * -1
        self.axis = random.randint(0, 180)
        self.add = Add(add=(random.choice(range(0, 250, 25)) / 100))
        self.vertical_prism = VerticalPrism(
            magnitude=random.randint(0, 5), direction=random.choice(["U", "D"])
        )
        self.horizontal_prism = HorizontalPrism(
            magnitude=random.randint(0, 5), direction=random.choice(["I", "O"])
        )

    def __str__(self) -> str:
        """Provide string representation of object."""

        def _give_plus_sign(value: float) -> str:
            """Will append '+' if the value is positive and will return value to 2 decimal places."""
            if value >= 0:
                return f"+{value:0.2f}"
            else:
                return f"{value:0.2f}"

        str_lst = []
        if self.sphere == 0:
            str_lst.append("plano")
        else:
            str_lst.append(_give_plus_sign(self.sphere))

        if self.sphere != 0 and self.cylinder == 0:
            str_lst.append(" DS")
        elif self.cylinder != 0:
            str_lst.append(" / ")
            str_lst.append(_give_plus_sign(self.cylinder))

            str_lst.append(" x ")
            str_lst.append(strip_decimal(self.axis))

        if self.add.add != 0:
            str_lst.append(" Add: ")
            str_lst.append(_give_plus_sign(self.add.add))
            str_lst.append(" @ ")
            str_lst.append(strip_decimal(self.add.working_distance_cm))
            str_lst.append("cm")

        return "".join(str_lst)



"""Supporting models for the `VisualAcuity` model."""

from pydantic import BaseModel as PydanticBaseModel


class BaseModel(PydanticBaseModel):
    """Monkey patching pydantic `BaseModel`."""

    class Config:
        """Configation of model."""

        validate_assignment = True



"""Custom exceptions related to the `prescrption` module."""

from typing import Any


class PrescriptionError(Exception):
    """Prescription input error."""

    def __init__(self, value: Any, message: str) -> None:
        """Construct exception."""
        self.value = value
        self.message = message
        super().__init__(message)



"""Functions for providing clean output."""


def strip_decimal(value: float, places: int = 1) -> str:
    """Remove a decimal if it is zero and returns the value rounded.

    Args:
        value (float): The value to be rounded.
        places (float): Number of places to round. Default is 1.

    Returns:
        (str): Rounded value with decimal stripped if equal to zero.
    """
    rounded_value = round(value, places)
    if (int(rounded_value) - rounded_value) == 0:
        return f"{int(rounded_value)}"
    return f"{rounded_value}"



"""Main entry point for optom_tools."""

from optom_tools.prescription import Prescription
from optom_tools.utils import log
from optom_tools.visual_acuity import VisualAcuity

__all__ = ["Prescription", "VisualAcuity", "log"]


def main():
    """Entry function."""
    pass


if __name__ == "__main__":
    main()



"""Supporting models for the main Prescription model."""

from typing import Optional

from pydantic import BaseModel as PydanticBaseModel
from pydantic import validator
from typing_extensions import Literal

from .exceptions import PrescriptionError


class BaseModel(PydanticBaseModel):
    """Monkey patching pydantic `BaseModel`."""

    class Config:
        """Configation of model."""

        validate_assignment = True


class BasePrism(BaseModel):
    """Base Prism model."""

    magnitude: float = 0

    @validator("magnitude", allow_reuse=True)
    @classmethod
    def _magnitude_validate(cls, value: float) -> float:
        """Validate magnitude.

        The magnitude of a prism must be positive.
        """
        if value < 0:
            raise PrescriptionError(
                value=value,
                message="The prism dioptre must be a positive number",
            )
        return value


class HorizontalPrism(BasePrism):
    """Horizontal Prism model.

    Note that the direction is related to base direction (e.g. right, left, in
    out).
    """

    direction: Optional[Literal["R", "L", "I", "O"]] = None


class VerticalPrism(BasePrism):
    """Vertical Prism model.

    Note that the direction is related to base direction (e.g. up, down).
    """

    direction: Optional[Literal["U", "D"]] = None


class Add(BaseModel):
    """Add model.

    A patient's add along with working distance (e.g. +2.00 @ 40cm for music).
    """

    add: float = 0.0
    working_distance_cm: float = 40
    description: str = ""

    @validator("add")
    @classmethod
    def _add_validate(cls, value: float) -> float:
        """Validate add.

        Add must be a positive number.
        """
        if value < 0:
            raise PrescriptionError(
                value=value, message="Add must be a positive number"
            )
        return value

    @validator("working_distance_cm")
    @classmethod
    def _working_distance_cm_validate(cls, value: float) -> float:
        """Validate add.

        The working distance must be positive and no more than 600cm (what is considered optical infinity).
        """
        if value < 0 or value > 600:
            raise PrescriptionError(
                value=value,
                message="Working Distance (cm) must be greater than 0 cm and no greater than 600cm",
            )
        return value



"""Main Visual Acuity class for module."""

import math
from decimal import Decimal
from typing import Optional, Tuple

import pydantic
from typing_extensions import Literal

from optom_tools.utils import strip_decimal

from .exceptions import VisualAcuityError
from .models import BaseModel

D = Decimal

FT_M = 0.3048  # 1 ft = 0.3048 m


class VisualAcuity(BaseModel):
    """The `visual_acuity` module contains methods in handling visual acuity measurements.

    Input for this class can be provided as keyword arguments. However, a string \
    such as '6/6' can be parsed as well.

    Args:
        numerator (int): The test distance.
        denominator (int): This distance required to subtend 5 minutes of arc.
        unit (Literal["ft", "m"]): Takes feet or metres/meters only. Defaults to 'm', but please check examples for nuances.

    Examples:
        Typical use:
        >>> va = VisualAcuity(numerator=6, denominator=6)
        >>> str(va)
        '6/6'

        If you want to explicitly mention the units:
        >>> va_feet = VisualAcuity(numerator=20, numerator=20, unit="ft")
        >>> str(va_feet)
        '20/20'

        Parsing more familiar input:
        >>> va_familiar = VisualAcuity('6/120')
        >>> va_familiar
        VisualAcuity(numerator=6.0, denominator=120.0, unit='m')

        If the numerator is greater `6`, then the unit is assumed to be feet.
        >>> va_familiar_ft = VisualAcuity('20/200')
        >>> va_familiar_ft
        VisualAcuity(numerator=20.0, denominator=200.0, unit='ft')
    """

    numerator: float = 6
    denominator: float = 6
    unit: Literal["ft", "m"] = "m"

    @pydantic.validator("numerator")
    @classmethod
    def _numerator_validator(cls, value):
        """Validate distance.

        Numerator must be a positive number.
        """
        if value < 0:
            raise VisualAcuityError(
                value=value, message="Distance must be a positive value"
            )
        return value

    @pydantic.validator("denominator")
    @classmethod
    def _denominator_validator(cls, value):
        """Validate denominator.

        Denominator must be a positive number.
        """
        if value < 0:
            raise VisualAcuityError(
                value=value, message="Distance must be a positive value"
            )
        return value

    def __init__(self, *args, **kwargs):
        """Init method."""
        if len(args) >= 1:
            components = self._simple_parse_va(args[0])
            kwargs["numerator"] = components[0]
            kwargs["denominator"] = components[1]
            kwargs["unit"] = components[2]
        super().__init__(**kwargs)

    def _convert_ft_m(self, ft: float) -> float:
        """Convert feet into metres/meters."""
        return round(ft * FT_M)

    def _convert_m_ft(self, m: float) -> float:
        """Convert meters/metres into feet."""
        return round(m / FT_M)

    def convert_unit(self, flag: Optional[Literal["ft", "m"]] = None) -> None:
        """Convert the unit of visual acuity between feet and metres.

        Args:
            flag (Optional[Literal["ft", "m"]]): Flag to ensure a conversion to a particular unit.

        Raises:
            VisualAcuityError: Only accepts `'ft'` and `'m'`as options.

        Examples:
            Converting without a flag:
            >>> va = VisualAcuity(numerator=6, denominator=6)
            >>> va.unit
            'm'
            >>> va.convert_unit()
            >>> str(va)
            '20/20'
            >>> va.unit
            'ft'

            Converting with an 'ft' flag:
            >>> va = VisualAcuity(numerator=20, denominator=20, unit="ft")
            >>> va.convert_unit("ft")
            >>> str(va)
            '20/20'
        """
        if flag not in ["ft", "m"]:
            raise VisualAcuityError(
                value=flag,
                message="Method convert_unit() only accepts flags 'ft' and 'm'",
            )
        if flag != self.unit:
            if self.unit == "ft":
                self.numerator = self._convert_ft_m(self.numerator)
                self.denominator = self._convert_ft_m(self.denominator)
                self.unit = "m"
            elif self.unit == "m":
                self.numerator = self._convert_m_ft(self.numerator)
                self.denominator = self._convert_m_ft(self.denominator)
                self.unit = "ft"

    @property
    def ft(self) -> str:
        """Return snellen fraction representation in feet.

        Returns:
            (str): Snellen Fraction in feet.

        Examples:
            Typical use:
            >>> va = VisualAcuity('6/6')
            >>> va.ft
            '20/20'
        """
        numerator = self.numerator
        denominator = self.denominator
        if self.unit != "ft":
            numerator = self._convert_m_ft(self.numerator)
            denominator = self._convert_m_ft(self.denominator)
        return f"{strip_decimal(numerator)}/{strip_decimal(denominator)}"

    @property
    def m(self) -> str:
        """Return snellen fraction in metres/meters.

        Returns:
            str: Snellen Fraction in meters/metres.

        Examples:
            Typical use:
            >>> va = VisualAcuity('20/20')
            >>> va.m
            '6/6'
        """
        numerator = self.numerator
        denominator = self.denominator
        if self.unit != "m":
            numerator = self._convert_ft_m(self.numerator)
            denominator = self._convert_ft_m(self.denominator)
        return f"{strip_decimal(numerator)}/{strip_decimal(denominator)}"

    @property
    def snellen_fraction(self) -> str:
        """Return snellen fraction representation of visual acuity.

        Returns:
            str: snellen fraction.

        Examples:
            Typical use:
            >>> VisualAcuity("6 / 12").snellen_fraction
            "6/6"
        """
        return f"{strip_decimal(self.numerator)}/{strip_decimal(self.denominator)}"

    @property
    def decimal(self) -> Decimal:
        """Return decimal form of visual acuity.

        Returns:
            Decimal: Decimal form of visual acuity.

        Examples:
            Typical use:
            >>> VisualAcuity("6/12").decimal
            0.5
        """
        return D(self.numerator / self.denominator)

    @property
    def logmar(self) -> Decimal:
        """Return logmar value of the visual acuity.

        Returns:
            Decimal: Logmar value of the visual acuity.

        Examples:
            Typical use:
            >>> VisualAcuity("6/6").logmar
            0.0
        """
        return D(math.log10(self.decimal))

    def _simple_parse_va(self, va: str) -> Tuple[str, str, Literal["ft", "m"]]:
        """Parse va string into a tuple.

        This tuple is numerator, denominator, unit.
        """
        components = va.split("/")
        if len(components) != 2:
            raise VisualAcuityError(
                value=va,
                message="Input must contain one '/' and numbers (e.g. '6/6')",
            )
        numerator = components[0]
        denominator = components[1]
        unit: Literal["ft", "m"] = "m"
        if numerator == "":
            numerator = "6"
        if float(numerator) > 6:
            unit = "ft"
        return (numerator, denominator, unit)

    def parse(self, va: str) -> BaseModel:
        """Parse a string into a visual acuity.

        Note: if the test distance (numerator) is greater than 6, then the unit is assumed to be in feet.

        Args:
            va (str): A visual acuity.

        Examples:
            Parsing a visual acuity:
            >>> va = VisualAcuity().parse('6/120')
            >>> str(va)
            '6/120'
            Parsing a visual acuity, automatically turning into a feet:
            >>> va = VisualAcuity().parse('20/20')
            >>> va.unit
            'ft'
        """
        components = self._simple_parse_va(va=va)
        self.numerator = float(components[0])
        self.denominator = float(components[1])
        self.unit = components[2]

        if self.numerator > 6:
            self.unit = "ft"

        return self

    def __str__(self) -> str:
        """Give string representation."""
        return self.snellen_fraction
